"""
Copyright 2008-2014 by Aerospike, Inc. All rights reserved.
THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE. THE COPYRIGHT NOTICE
ABOVE DOES NOT EVIDENCE ANY ACTUAL OR INTENDED PUBLICATION.
"""


# -*- coding: utf-8 *-*


import functools
import json
from libs import base
from flask import current_app
from flask_kvsession import KVSessionExtension
from simplekv.memory import DictStore


store = DictStore()
KVSessionExtension(store, current_app)


class SessionManager(object):

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(SessionManager, cls).__new__(
                                cls, *args, **kwargs)
            cls._instance.__init__()
        return cls._instance

    def __init__(self):
        self._sessions = dict()

    def create_session(self, session_key):
        self._sessions[session_key] = dict()

    def delete_session(self, session_key):

        #session['session'] = ''
        import sys
        import time
        import datetime
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
        sys.stderr.write('\n'+ str(st) + ' : Logging Out.....')
        del(self._sessions[session_key])

    def get_clusters_from_session(self, session_key):

        list_dict = []
        if session_key not in self._sessions:
            raise Exception('Invalid Session ID')
        try:
            for cluster_id in self._sessions[session_key]:
                temp_dict = dict()
                temp_dict['username'] = self._sessions[session_key][cluster_id]['user']
                temp_dict['cluster_name'] = self._sessions[session_key][cluster_id]['cluster_name']
                roles = self._sessions[session_key][cluster_id]['roles']
                temp_dict['roles'] = roles if roles is None else list(roles)
                temp_dict['seed_node'] = self._sessions[session_key][cluster_id]['seed_node']
                temp_dict['cluster_id'] = cluster_id
                list_dict.append(temp_dict)
        except:
            raise Exception('Invalid Cluster ID')
        return list_dict

        """temp_dict = self._sessions[session_key]
        for key in self._sessions[session_key]:
            for cluster_key in self._sessions[session_key][key]:
                del temp_dict[session_key][key][cluster_key]['password']
                del temp_dict[session_key][key][cluster_key]['roles']
        return temp_dict"""

    def add_cluster(self, session_key, node_address, cluster_id, cluster_name, user=None, password=None, pl_password=None):

        ip, port = node_address.split(":")
        build = '0.0.0'
        if user is not None:
            build = base.get_node_build(ip, port, user=user, password=password)
        user_roles = base.get_cluster_userroles(ip, port,user=user, password=password)
        if build > '3.5.4' or build == '3.5.4':
            user_privileges = []
            for role in user_roles:
                status, privileges = base.query_role(ip, port, user, password, role)
                for privilege in privileges:
                    user_privileges.append(privilege.split('.')[0])
            user_roles = user_privileges
            user_roles = list(set(user_roles))
            #user_roles = ['read-write' if role=='read-write-udf' else role for role in user_roles]
        if user_roles is not None:
            user_roles = set(user_roles)
        self._sessions[session_key][cluster_id] = dict(user= user, password= password,
             roles = user_roles, cluster_name = cluster_name,
             seed_node = node_address, pl_password = pl_password)

    def remove_cluster(self, session_key, cluster_id):

        del(self._sessions[session_key][cluster_id])
        if cluster_id not in self._sessions[session_key]:
            return True
        return False

    def get_user_roles_by_cluster_id(self, session_key, cluster_id):
        temp_dict = self._sessions[session_key][cluster_id]
        return temp_dict['roles']

    def are_roles_matching(self, session_key, cluster_id, roles):

        user_roles = self.get_user_roles_by_cluster_id(session_key, cluster_id)
        if user_roles is None:
            return True
        common_roles = user_roles.intersection(roles)
        return True if len(common_roles) != 0 else False

    def get_pl_password(self, session_key, cluster_id):
        temp_dict = self._sessions[session_key][cluster_id]
        return temp_dict['pl_password']

    def is_valid_session(self, session_key):
        return session_key in self._sessions

    def is_cluster_present(self, session_key, node_address=None, cluster_name=None, search_node_list=False, cls_mgr=None):

        if node_address is not None:
            for cluster_id, data in self._sessions[session_key].items():
                if data['seed_node'] == node_address:
                    if cluster_name is not None:
                        self._sessions[session_key][cluster_id]['cluster_name'] = cluster_name
                    return cluster_id
        elif cluster_name is not None:
            for cluster_id, data in self._sessions[session_key].items():
                if data['cluster_name'] == cluster_name:
                    return cluster_id
        if search_node_list and node_address is not None:
            for cluster_id, data in self._sessions[session_key].items():
                if data['seed_node'] != node_address and cls_mgr is not None:
                    cluster_data = cls_mgr.get_cluster_data_by_cluster_id(cluster_id)
                    if 'nodes' in cluster_data and len(cluster_data['nodes']) != 0:
                        if node_address in cluster_data['nodes']:
                            return cluster_id
        return False

    def is_cluster_id_present(self, session_key, cluster_id):

        if cluster_id in self._sessions[session_key]:
            return True
        return False

    def get_user(self, session_id, cluster_key):

        #if cluster_key in self._sessions[session_id]:
            #sys.stderr.write('\ncluster key present!! ' + str(cluster_key) + '\n' + str(self._sessions[session_id][cluster_key]))
        #else:
            #sys.stderr.write('\ncluster key ABSENT!!')
        return self._sessions[session_id][cluster_key]['user']

    def get_cluster_name(self, session_id, cluster_key):

        return self._sessions[session_id][cluster_key]['cluster_name']

    def update_cluster_name(self, session_key, cluster_id, cluster_name):

        self._sessions[session_key][cluster_id]['cluster_name'] = cluster_name

    def update_cluster_id(self, session_key, old_cluster_id, new_cluster_id):

        temp_dict = self._sessions[session_key][old_cluster_id]
        self._sessions[session_key][new_cluster_id] = temp_dict
        del(self._sessions[session_key][old_cluster_id])

    def remove_cluster_from_session_containing_user(self, username, seed):

        temp_sessions = {}
        #seed_address = None
        #cluster_id = None
        response = []
        for session in self._sessions:
            temp_sessions[session] = dict()
            for cluster in self._sessions[session]:
                if self._sessions[session][cluster]['user'] != username:
                    temp_sessions[session][cluster] = self._sessions[session][cluster]
                else:
                    if str(self._sessions[session][cluster]['seed_node']) == str(seed):
                        entry = dict()
                        entry['seed_node'] = self._sessions[session][cluster]['seed_node']
                        entry['cluster_id'] = cluster
                        response.append(entry)
                    #seed_address = self._sessions[session][cluster]['seed_node']
                    #cluster_id = cluster


        self._sessions = temp_sessions
        return response
        #return str(seed_address), cluster_id
        """t_sessions = {}

        for session in temp_sessions:
            if len(temp_sessions[session]) != 0:
                t_sessions[session] = temp_sessions[session]

        sys.stderr.write('\nt_sessions: '+ str(t_sessions))
        self._sessions = t_sessions"""

    def change_password(self, session_key, cluster_id, password, pl_password):

        self._sessions[session_key][cluster_id]['password'] = str(password)
        self._sessions[session_key][cluster_id]['pl_password'] = str(pl_password)

    def remove_cluster_from_other_sessions(self, session_key, user, node_address):

        temp_sessions = {}
        response = []
        for session in self._sessions:
            temp_sessions[session] = dict()
            if session == session_key:
                temp_sessions[session]= self._sessions[session]
                continue
            for cluster in self._sessions[session]:
                if self._sessions[session][cluster]['user'] == user and self._sessions[session][cluster]['seed_node'] == node_address:
                    entry = dict()
                    entry['seed_node'] = self._sessions[session][cluster]['seed_node']
                    entry['cluster_id'] = cluster
                    response.append(entry)
                else:
                    temp_sessions[session][cluster] = self._sessions[session][cluster]

        self._sessions = temp_sessions
        return response

    def get_all_sessions(self):

        return self._sessions


    def check_session_validity(self, session_obj, roles=None):

        def decorator(fn):

            @functools.wraps(fn)
            def f(*args, **kwargs):

                """if roles is not None:
                        import sys
                        sys.stderr.write('\nRoles are: ' + str(roles))"""
                session_key = session_obj.get('session', None)
                cluster_id = kwargs.get('cluster', None)
                if session_key is not None and cluster_id is None:
                    return fn(*args, **kwargs)
                elif session_key is not None and cluster_id is not None:
                    if session_key not in self._sessions:
                        """This condition handles sesssion key removed on user update from user mgmt."""
                        return json.dumps(dict(error="Invalid session: %s" % session_key))
                    if cluster_id not in self._sessions[session_key]:
                        return json.dumps(dict(error="Invalid cluster id : %s" % kwargs["cluster"]))
                    temp_dict = self._sessions[session_key][cluster_id]
                    kwargs['user'] = temp_dict['user']
                    kwargs['password'] = temp_dict['password']

                    if roles is None:
                        return fn(*args, **kwargs)
                    elif self.are_roles_matching(session_key, cluster_id, roles):
                        return fn(*args, **kwargs)
                    else:
                        return json.dumps(dict(
                            error="You do not have enough permissions to access it."))
                elif session_key is None:
                    return json.dumps(dict(error="Invalid session : %s" % session_key))
                elif cluster_id is None:
                    return json.dumps(dict(error="Invalid cluster id : %s" % kwargs["cluster"]))

            return f
        return decorator



